
#![feature(dataclasses)]
#![feature(dict_ordered)]
#![feature(fstrings)]
#![feature(member_annotations)]

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '0.2.0'

from ._converter import *
from ._datamodel import *
from ._locator import Locator
from ._typing import *
from ._union import *




from typing import Callable, Iterable, Optional, TypeVar
from typing import Any, Iterator, Optional, MutableMapping, TypeVar, Union, overload
from collections import abc

__all__ = ['find', 'ChainDict']
T = TypeVar('T')
KT = TypeVar('KT')
VT = TypeVar('VT')


def find(predicate: Callable[[Optional[T]], bool], it: Iterable[T]) -> Optional[T]:
  return next(filter(predicate, it), None)


class ChainDict(MutableMapping[KT, VT]):
  """
  A dictionary that wraps a list of dictionaries. The dictionaries passed
  into the #ChainDict will not be mutated. Setting and deleting values will
  happen on the first dictionary passed.
  """

  def __init__(self, *dicts):
    if not dicts:
      raise ValueError('need at least one argument')
    self._major = dicts[0]
    self._dicts = list(dicts)
    self._deleted = set()
    self._in_repr = False

  def __repr__(self) -> str:
    if self._in_repr:
      return 'ChainDict(...)'
    else:
      self._in_repr = True
      try:
        return 'ChainDict({})'.format(dict(self.items()))
      finally:
        self._in_repr = False

  def __iter__(self) -> Iterator[KT]:
    seen = set()
    for d in self._dicts:
      for key in d.keys():
        if key not in seen and key not in self._deleted:
          yield key
          seen.add(key)

  def __len__(self) -> int:
    return sum(1 for x in self.keys())

  def __contains__(self, key: Any) -> bool:
    if key not in self._deleted:
      for d in self._dicts:
        if key in d:
          return True
    return False

  def __getitem__(self, key: KT) -> VT:
    if key not in self._deleted:
      for d in self._dicts:
        try: return d[key]
        except KeyError: pass
    raise KeyError(key)

  def __setitem__(self, key: KT, value: VT) -> None:
    self._major[key] = value
    self._deleted.discard(key)

  def __delitem__(self, key: KT) -> None:
    if key not in self:
      raise KeyError(key)
    self._major.pop(key, None)
    self._deleted.add(key)

  def clear(self):
    self._major.clear()
    self._deleted.update(self.keys())




import abc
from dataclasses import dataclass as _dataclass
from typing import Any, Dict, Generic, List, Mapping, Optional, Type, TypeVar, Union
from ._datamodel import (
  BaseMetadata,
  FieldMetadata,
  ModelMetadata,
  UnionMetadata,
  datamodel,
  uniontype,
)
from ._locator import Locator
from ._typing import type_repr
from .utils import ChainDict

T = TypeVar('T')

__all__ = [
  'ConversionError',
  'ConversionTypeError',
  'ConversionValueError',
  'Context',
  'Converter',
  'UnknownTypeError',
  'Registry',
]


@_dataclass
class ConversionError(Exception):
  context: 'Context'
  message: str

  def __str__(self):
    return '{}: {}'.format(self.context.locator, self.message)


class ConversionTypeError(ConversionError, TypeError):
  pass


class ConversionValueError(ConversionError, ValueError):
  pass


@_dataclass
class Context:
  """
  The context contains data relevant during value conversion, such as metadata coming from
  union types, data models and fields. The context represents a tree, where every element in
  that tree represents a value in the structured data that is being converted.
  """

  parent: Optional['Context']
  registry: 'Registry'
  locator: Locator
  type: Type
  value: Any
  field_metadata: Optional[FieldMetadata]

  @classmethod
  def new(cls, registry: 'Registry', type_: Type, value: Any, field_metadata: FieldMetadata=None) -> 'Context':
    return cls(None, registry, Locator([]), type_, value, field_metadata)

  def fork(
    self,
    type_: Type,
    value: Any,
    field_metadata: Optional[FieldMetadata] = NotImplemented,
  ) -> 'Context':
    """
    Create a fork in the context tree, re-using the same locator and parent but allowing to
    change the type and value, and optionally the field metadata.
    """

    if field_metadata is NotImplemented:
      field_metadata = self.field_metadata

    return Context(self.parent, self.registry, self.locator, type_, value, field_metadata)

  def child(
    self,
    key: Union[int, str],
    type_: Type,
    value: Any,
    field_metadata: FieldMetadata=None
  ) -> 'Context':
    """
    Create a new child node in the context tree, advancing to the next sub-structure from the
    current value.
    """

    return Context(self, self.registry, self.locator.push(key), type_, value, field_metadata)

  def get_converter(self) -> 'Converter':
    return self.registry.get_converter(self.type)

  def from_python(self) -> Any:
    return self.get_converter().from_python(self.value, self)

  def to_python(self) -> Any:
    return self.get_converter().to_python(self.value, self)

  def type_error(self, message: str) -> ConversionTypeError:
    return ConversionTypeError(self, message)

  def value_error(self, message: str) -> ConversionValueError:
    return ConversionValueError(self, message)



class Converter(Generic[T], metaclass=abc.ABCMeta):
  """
  Abstract base class that convert from and to a Python datatype.
  """

  @abc.abstractmethod
  def from_python(self, value: T, context: Context) -> Any:
    pass

  @abc.abstractmethod
  def to_python(self, value: Any, context: Context) -> T:
    pass


class UnknownTypeError(TypeError):
  pass


class Registry:
  """
  The registry is what maps data types and type hints to #Converter implementations. Types that
  are decorated with #@uniontype() or #@datamodel() are handled special in that they must be
  associated with the respective decorator function.
  """

  def __init__(self, parent: Optional['Registry']) -> None:
    self.parent = parent
    self._mapping: Dict[Any, Converter] = {}
    self._type_options: Dict[Any, Dict[str, Any]] = {}

  @property
  def root(self) -> 'Registry':
    if not self.parent:
      return self
    return self.parent.root

  def register_converter(self, type_: Any, converter: Converter, overwrite: bool = False) -> None:
    """
    Registers a convert for the specified Python type or type hint.
    """

    old = type_
    type_ = normalize_type(type_, keep_parametrized=True)
    if type_ in self._mapping and not overwrite:
      raise RuntimeError(f'converter for {type_repr(type_)} already registered')
    if not isinstance(converter, Converter):
      raise TypeError(f'expected Converter, got {type_repr(type(converter))}')
    self._mapping[type_] = converter

  def update_options(self, type_: Any, options: Dict[str, Any]) -> None:
    """
    Registers options with the specified Python type or type hint. Existing options are
    merged, with the options specified to this method taking precedence.
    """

    self._type_options.setdefault(type_, {}).update(options)

  def set_option(self, type_: Any, option_name: str, value: Any) -> None:
    """
    Set a specific option.
    """

    self._type_options.setdefault(type_, {})[option_name] = value

  def get_options(self, type_: Any) -> Mapping[str, Any]:
    """
    Returns a mapping that contains all options for the specified type or type hint, taking
    options defined on the parent #Registry into account.

    Note that this does not respect type inheritance.
    """

    options: Mapping[str, Any] = self._type_options.get(type_, {})
    if self.parent:
      options = ChainDict(options, self.parent.get_options(type_))
    return options

  def get_option(self, type_: Any, option_name: str, default: Any = None) -> Any:
    """
    Return a specific option associated with the specified type or type hint.
    """

    return self.get_options(type_).get(option_name, default)

  def get_converter(self, type_: Type) -> Converter:
    """
    Return a converter registered for this type or type hint. If there is no immediate match
    for the type, it will be normalized using #normalize_type().
    """

    if type_ in self._mapping:
      return self._mapping[type_]

    if self.parent:
      try:
        return self.parent.get_converter(type_)
      except UnknownTypeError:
        pass

    # Try base classes.
    for base in getattr(type_, '__bases__', ()):
      try:
        return self.get_converter(base)
      except UnknownTypeError:
        pass

    normalized = normalize_type(type_, keep_parametrized=False)
    if normalized != type_:
      return self.get_converter(normalized)

    raise UnknownTypeError(f'no converter found for type {type_repr(type_)}')

  def make_context(self, type_: Type, value: Any, field_metadata: FieldMetadata = None) -> Any:
    return Context.new(self, type_, value, field_metadata)


def normalize_type(type_: Any, keep_parametrized: bool) -> Type:
  """
  Normalizes a Python type or type hint. For type hints, this will return the `__origin__`.
  If *keep_parametrized* is `True`, then the `__origin__` will only be returned if the type
  hint is not parametrized (i.e. still generic).
  """

  # Map type's decoreated with uniontype/datamodel to the respective functions.
  metadata = BaseMetadata.for_type(type_)
  if isinstance(metadata, UnionMetadata):
    type_ = uniontype
  elif isinstance(metadata, ModelMetadata):
    type_ = datamodel

  # Resolve type hints to the original annotated form.
  # NOTE: In Python 3.6, Dict.__origin__ is None whereas in Python 3.7 it is dict.
  if getattr(type_, '__origin__', None) and (not keep_parametrized or type_.__parameters__):
    type_ = type_.__origin__

  return type_




import string
from typing import Any, Dict, List, Sequence, Union


class Locator:
  """
  A locator represents the path of an object in a nested data structure as a list of string and
  integer indices. Locator objects are immutable.
  """

  _WHITELISTED_KEY_CHARACTERS = string.ascii_letters + string.digits + '_-'

  def __init__(self, items: Sequence[Union[str, int]]) -> None:
    self._items = items

  __iter__ = lambda self: iter(self._items)
  __len__ = lambda self: len(self._items)
  __getitem__ = lambda self, index: self._items[index]
  __repr__ = lambda self: 'Locator({})'.format(str(self))
  __bool__ = lambda self: bool(self._items)
  __nonzero__ = __bool__

  def __str__(self):
    parts = ['$']
    for item in self._items:
      if isinstance(item, int):
        parts.append('[' + str(item) + ']')
      else:
        item = str(item)
        if '"' in item:
          item = item.replace('"', '\\"')
        if any(c not in self._WHITELISTED_KEY_CHARACTERS for c in item):
          item = '"' + item + '"'
        parts.append('.' + item)
    return ''.join(parts)

  def resolve(self, root: Union[Any, Dict, List]) -> Any:
    """
    Accesses the value represented by this Locator object starting from *root*.

    >>> root = {'values': [1, 2, 3]}
    >>> Locator(['values', 1]).access(root)
    2

    :param root: The object to start indexing from.
    :raises KeyError: If an item in an object cannot be accessed.
    :raises IndexError: If an item in a list cannot be accessed.
    ``` """

    value = root
    for item in self._items:
      try:
        value = value[item]  # type: ignore
      except KeyError as exc:
        raise KeyError(str(self))
      except IndexError as exc:
        raise IndexError('{} at {}'.format(exc, self))

    return value

  def push(self, item: Union[str, int]) -> 'Locator':
    """
    Create a copy of this locator with *item* appended to it.
    """

    items = list(self._items)
    items.append(item)
    return Locator(items)

  def pop(self) -> 'Locator':
    """
    Creates a copy of this locator, with the last element removed.
    """

    items = list(self._items)
    items.pop()
    return Locator(items)




"""
Utilities for type hints not provided by the #typing module.
"""

import types
from typing import Any

__all__ = [
  'type_repr',
]


def _type_repr(obj: Any) -> str:
  # Borrowed from typing in CPython 3.7
  if isinstance(obj, type):
      if obj.__module__ == 'builtins':
          return obj.__qualname__
      return f'{obj.__module__}.{obj.__qualname__}'
  if obj is ...:
      return('...')
  if isinstance(obj, types.FunctionType):
      return obj.__name__
  return repr(obj)


try:
  from typing import _type_repr as type_repr  # type: ignore
except ImportError:
  type_repr = _type_repr




import abc
import dataclasses
from typing import Dict, List, Type, Union
from ._typing import type_repr

__all__ = [
  'UnionTypeError',
  'UnionResolver',
]


class UnionTypeError(TypeError):
  pass


class UnionResolver(metaclass=abc.ABCMeta):
  """
  Abstract base class for union type resolvers.
  """

  @abc.abstractmethod
  def type_for_name(self, type_name: str) -> Type:
    """
    Returns the type for the specified *type_name*. Raise #UnionTypeError if the type name
    cannot be resolved.
    """

  @abc.abstractmethod
  def name_for_type(self, type: Type) -> str:
    """
    Returns the name for the specified *type* (which could be used to retrieve the same type
    with #type_for_name()). Raise #UnionTypeError if the type cannot be named.
    """

  @abc.abstractmethod
  def members(self) -> List[str]:
    """
    Enumerate the union members by name. Implementations may raise a #NotImplementedError.
    """


class StaticUnionResolver(UnionResolver):
  """
  Wraps a dictionary for resolving union members.
  """

  def __init__(self, mapping: Dict[str, Type]) -> None:
    self._mapping = dict(mapping)

  def __getitem__(self, type_name: str) -> Type:
    return self._mapping[type_name]

  def __setitem__(self, type_name: str, type_: Type) -> None:
    self._mapping[type_name] = type_

  def __delitem__(self, type_name: str) -> None:
    del self._mapping[type_name]

  def __eq__(self, other):
    if not isinstance(other, StaticUnionResolver):
      return False
    return self._mapping == other._mapping

  def __ne__(self, other):
    if not isinstance(other, StaticUnionResolver):
      return True
    return self._mapping != other._mapping

  def type_for_name(self, type_name: str) -> Type:
    try:
      return self._mapping[type_name]
    except KeyError:
      raise UnionTypeError(f'type name {type_name!r} could not be resolved')

  def name_for_type(self, type_: Type) -> str:
    for key, value in self._mapping.items():
      if type_ is value:
        return key
      # Check for Optional[type_]
      if (getattr(value, '__origin__', None) == Union and len(value.__args__) == 2 and
          value.__args__[1] == type(None) and type_ == value.__args__[0]):
        return key
    raise UnionTypeError(f'type {type_repr(type_)} could not be resolved')

  def members(self) -> List[str]:
    return list(self._mapping.keys())




"""
Extends the functionality of the #dataclass module to provide additional metadata for
(de-) serializing data classes.
"""

import types
import textwrap
from typing import Any, Dict, Iterable, List, Optional, Union, Tuple, Type, TypeVar, cast, get_type_hints

from dataclasses import dataclass as _dataclass, field as _field, Field as _Field, _MISSING_TYPE

from ._typing import type_repr
from ._union import UnionResolver, StaticUnionResolver

T = TypeVar('T')
T_BaseMetadata = TypeVar('T_BaseMetadata', bound='BaseMetadata')

__all__ = [
  'UnionMetadata',
  'ModelMetadata',
  'FieldMetadata',
  'uniontype',
  'datamodel',
  'field',
  'enumerate_fields',
  'is_datamodel',
  'is_uniontype',
]


def _extract_dataclass_from_kwargs(dataclass: Type[T], kwargs: Dict[str, Any]) -> T:
  """
  Internal. Extracts all keyword arguments matching the fields in *dataclass* from the
  dictionary *kwargs* and returns an instance of *dataclass*.
  """

  return dataclass(**{  # type: ignore
    field.name: kwargs.pop(field.name)
    for field in dataclass.__dataclass_fields__.values()  # type: ignore
    if field.name in kwargs
  })


def _field_has_default(field: _Field) -> bool:
  return any(not isinstance(x, _MISSING_TYPE) for x in (field.default, field.default_factory))  # type: ignore


def _field_get_default(field: _Field) -> Any:
  if not isinstance(field.default, _MISSING_TYPE):  # type: ignore
    return field.default
  if not isinstance(field.default_factory, _MISSING_TYPE):  # type: ignore
    return field.default_factory()  # type: ignore
  raise RuntimeError('{!r} has no default'.format(field))


@_dataclass
class BaseMetadata:
  #: Use strict deserialization mode where that is not already the default.
  strict: bool = _field(default=False)

  #: Use relaxed deserialization mode where that is not already the default.
  relaxed: bool = _field(default=False)

  ATTRIBUTE = '__databind_metadata__'

  @classmethod
  def for_type(cls: Type[T_BaseMetadata], type_: Type) -> T_BaseMetadata:
    try:
      result = vars(type_).get(cast(BaseMetadata, cls).ATTRIBUTE)
    except TypeError:
      return cls()
    if result is None or not isinstance(result, cls):
      result = cls()
    return result


@_dataclass
class ModelMetadata(BaseMetadata):
  #: Allow only keyword arguments when constructing an instance of the model.
  kwonly: bool = _field(default=False)


@_dataclass
class FieldMetadata(BaseMetadata):
  #: Alternative name of the field used during the (de-) serialization. This may be
  #: set to change the name of a field to a value that is not valid Python syntax
  #: (for example "field-name").
  altname: Optional[str] = _field(default=None)

  #: Set to `True` to indicate that the field is required during deserialization.
  #: This does not indicate that the field is required when constructing the object
  #: in memory, in which case the field's default value is always used.
  required: bool = _field(default=False)

  #: Set to `True` if the field is populated with derived data and should be ignored
  #: during (de-) serialization.
  derived: bool = _field(default=False)

  #: Set to `True` if the type of the field should be flattened into the parent
  #: object when (de-) serializing. There can be only one flattened field on a data
  #: model and it works with other data models as well as dictionaries.
  flatten: bool = _field(default=False)

  #: Indicate that the field should be assigned the raw data from the deserialization.
  #: Setting this to `True` sets #derived to `True` as well.
  raw: bool = _field(default=False)

  #: A list of format specifiers that the (de-) serializer should respect, if applicable
  #: to the field type. Examples of format specifiers would be date format templates for
  #: date types or #decimal.Context for #decimal.Decimal fields.
  formats: List[Any] = _field(default_factory=list)

  #: The original #_Field that this #FieldMetadata is associated with. This may not be set
  #: if the #FieldMetadata is instantiated at the top level. NOTE: We should use a weak
  #: reference here, but the #_Field class is not compatible with weakrefs.
  _owning_field: 'Optional[_Field]' = _field(default=None)

  #: Metadata that supplements the original #_Field's metadata. When a #FieldMetadata is
  #: instantiated for a field in a #datamodel, the metadata will be set to the metadata
  #: of the #_Field.
  metadata: Dict[str, Any] = _field(default_factory=dict)

  KEYSPACE = 'databind.core'

  def __post_init__(self):
    if self.raw:
      self.derived = True

  @classmethod
  def for_field(cls, field: _Field) -> 'FieldMetadata':
    return field.metadata.get(cls.KEYSPACE) or cls(_owning_field=field)


def datamodel(*args, **kwargs):
  """
  This function wraps the #dataclasses.dataclass() decorator. Applicable keyword arguments are
  redirected to the #ModelMetadata which is then set in the `__databind_metadata__` attribute.
  """

  metadata = _extract_dataclass_from_kwargs(ModelMetadata, kwargs)
  uninitialized = object()  # Placeholder object to inidicate that a field does not actually have a default.

  def _before_dataclass(cls):
    # Allow non-default arguments to follow default-arguments.
    no_default_fields = []
    for key, value in getattr(cls, '__annotations__', {}).items():
      if not hasattr(cls, key):
        f = field()
        setattr(cls, key, f)
      else:
        f = getattr(cls, key)
        if not isinstance(f, _Field):
          continue
      if not _field_has_default(f):
        # This prevents a SyntaxError if non-default arguments follow default arguments.
        f.default = uninitialized
        no_default_fields.append(key)

    # Override the `__post_init__` method that is called by the dataclass `__init__`.
    old_postinit = getattr(cls, '__post_init__', None)
    def __post_init__(self):
      # Ensure that no field has a "uninitialized" value.
      for name in no_default_fields:
        if getattr(self, name) == uninitialized:
          raise TypeError(f'missing required argument {name!r}')
      if old_postinit:
        old_postinit(self)
    cls.__post_init__ = __post_init__

    return cls

  def _after_dataclass(cls):
    setattr(cls, ModelMetadata.ATTRIBUTE, metadata)

    if metadata.kwonly:
      old_init = cls.__init__
      def __init__(self, **kwargs):
        old_init(self, **kwargs)
      cls.__init__ = __init__

    return cls

  if args:
    _before_dataclass(args[0])

  result = _dataclass(*args, **kwargs)

  if not args:
    def wrapper(cls):
      return _after_dataclass(result(_before_dataclass(cls)))
    return wrapper

  return _after_dataclass(result)


def field(*args, **kwargs) -> _Field:
  """
  This function wraps the #dataclasses.field() function, equipping the field with additional
  #FieldMetadata.
  """

  metadata = kwargs.pop('metadata', {})
  field_metadata = _extract_dataclass_from_kwargs(FieldMetadata, kwargs)
  metadata[FieldMetadata.KEYSPACE] = field_metadata
  kwargs['metadata'] = metadata
  field = _field(*args, **kwargs)
  field_metadata._owning_field = field
  field_metadata.metadata = field.metadata
  return field


@_dataclass
class _EnumeratedField:
  field: _Field = _field(repr=False)
  name: str
  type: Type
  metadata: FieldMetadata = _field(repr=False)

  def has_default(self) -> bool:
    return _field_has_default(self.field)

  def get_default(self) -> Any:
    return _field_get_default(self.field)


def enumerate_fields(data_model: Union[T, Type[T]]) -> Iterable[_EnumeratedField]:
  """
  Enumerate the fields of a datamodel. The items yielded by this generator are tuples of
  the field name, the resolved type hint and the #FieldMetadata.
  """

  if not isinstance(data_model, type):
    data_model = type(data_model)

  type_hints = get_type_hints(data_model)
  for field in data_model.__dataclass_fields__.values():  # type: ignore
    yield _EnumeratedField(field, field.name, type_hints[field.name], FieldMetadata.for_field(field))


def is_datamodel(obj: Any) -> bool:
  return isinstance(BaseMetadata.for_type(obj), ModelMetadata)


@datamodel
class UnionMetadata(BaseMetadata):
  #: The resolver that is used to map type names to a datatype. If no resolver is set,
  #: the union type acts as a container for exactly one of it's dataclass fields.
  resolver: UnionResolver

  #: Indicates whether the union type is a container for its members.
  container: bool = _field(default=False)

  #: The type field of the union container.
  type_field: Optional[str] = _field(default=None)

  #: The key in the data structure that identifies the union type. Defaults to "type".
  type_key: str = _field(default='type')

  #: Whether union members should be converted from Python to flat data strucutres.
  #: This option is required in order to de-serialize union members that cannot be
  #: deserialized from mappings (like plain types). The default is `True`.
  flat: bool = _field(default=True)


def uniontype(
  resolver: Union[UnionResolver, Dict[str, Type], Type] = None,
  container: bool = False,
  type_field: str = None,
  **kwargs
):
  """
  Decorator for classes to define them as union types, and are to be (de-) serialized as such.
  Union types can either act as placeholders or as containers for their members.

  If a *resolver* is specified or *container* is `False`, the union type will act as a placeholder
  and will be converted directly into one the union member types. If no *resolver* is specified,
  the members will be determined from the type hints.

  If *container* is #True, the union type will be equipped as a container for its members
  based on the type hints. The *type_field* will be used to store the type name in the
  container.
  """

  cls = None
  if isinstance(resolver, type):
    resolver, cls = None, resolver
    #import pdb; pdb.set_trace()

  if container:
    if not type_field:
      type_field = kwargs.get('type_key', 'type')
    kwargs.setdefault('type_key', type_field)

  if resolver is not None:
    if container:
      raise TypeError('"container" argument cannot be combined with "resolver" argument')
    if type_field is not None:
      raise TypeError('"type_field" argument cannot be combined with "resolver" argument')
    if isinstance(resolver, dict):
      resolver = StaticUnionResolver(resolver)

  def _init_as_container(cls, type_hints):
    """
    Initializes *cls* as a container for the union type members.
    """

    scope = {}
    exec(textwrap.dedent(f"""
      def __init__(self, {type_field}: str, value: Any) -> None:
        self.{type_field} = {type_field}
        self._value = value
      def __repr__(self) -> str:
        return f'{{type(self).__name__}}({{self.{type_field}!r}}, {{self._value!r}})'
      def __eq__(self, other) -> bool:
        if isinstance(other, cls):
          return self.{type_field} == other.{type_field} and self._value == other._value
        return False
      def __ne__(self, other) -> bool:
        if isinstance(other, cls):
          return self.{type_field} != other.{type_field} or self._value != other._value
        return True
    """), {'Any': Any, 'cls': cls}, scope)

    for key, value in scope.items():
      if key not in vars(cls):
        setattr(cls, key, scope[key])

    def _make_property(type_name: str, annotation: Any) -> property:
      def getter(self) -> annotation:  # type: ignore
        assert type_field is not None
        has_type = getattr(self, type_field)
        if has_type != type_name:
          raise TypeError(f'{type_repr(cls)}.{type_name} cannot be accessed if {type_field} == {has_type!r}')
        return self._value
      def setter(self, value: annotation) -> None:  # type: ignore
        assert type_field is not None
        setattr(self, type_field, type_name)
        self._value = value
      return property(getter, setter)

    for key, annotation in type_hints.items():
      setattr(cls, key, _make_property(key, annotation))

    return cls

  def _prevent_init(cls):
    def __init__(self, *args, **kwargs):
      raise TypeError(f'non-container @uniontype {type_repr(cls)} cannot be constructed directly')
    if '__init__' not in vars(cls):
      cls.__init__ = __init__

  def decorator(cls):
    nonlocal resolver
    if not resolver or container:
      type_hints = get_type_hints(cls)
    if not resolver:
      resolver = StaticUnionResolver(type_hints)
    if container:
      _init_as_container(cls, type_hints)
    else:
      _prevent_init(cls)

    kwargs['resolver'] = resolver
    kwargs['container'] = container
    kwargs['type_field'] = type_field
    setattr(cls, UnionMetadata.ATTRIBUTE, UnionMetadata(**kwargs))

    return cls

  if cls:
    return decorator(cls)

  return decorator


def is_uniontype(obj: Any) -> bool:
  return isinstance(BaseMetadata.for_type(obj), UnionMetadata)
